import type { Frame } from './types/Frame';
import type { IWorkletContext } from 'react-native-worklets-core';
declare global {
    var __frameProcessorRunAtTargetFpsMap: Record<string, number | undefined> | undefined;
    var __ErrorUtils: {
        reportFatalError: (error: unknown) => void;
    } | undefined;
}
type BasicParameterType = string | number | boolean | undefined;
type ParameterType = BasicParameterType | BasicParameterType[] | Record<string, BasicParameterType | undefined>;
/**
 * An initialized native instance of a FrameProcessorPlugin.
 * All memory allocated by this plugin will be deleted once this value goes out of scope.
 */
export interface FrameProcessorPlugin {
    /**
     * Call the native Frame Processor Plugin with the given Frame and options.
     * @param frame The Frame from the Frame Processor.
     * @param options (optional) Additional options. Options will be converted to a native dictionary
     * @returns (optional) A value returned from the native Frame Processor Plugin (or undefined)
     */
    call(frame: Frame, options?: Record<string, ParameterType>): ParameterType;
}
interface TVisionCameraProxy {
    /**
     * @internal
     */
    setFrameProcessor(viewTag: number, frameProcessor: (frame: Frame) => void): void;
    /**
     * @internal
     */
    removeFrameProcessor(viewTag: number): void;
    /**
     * Creates a new instance of a native Frame Processor Plugin.
     * The Plugin has to be registered on the native side, otherwise this returns `undefined`.
     * @param name The name of the Frame Processor Plugin. This has to be the same name as on the native side.
     * @param options (optional) Options, as a native dictionary, passed to the constructor/init-function of the native plugin.
     * @example
     * ```ts
     * const plugin = VisionCameraProxy.initFrameProcessorPlugin('scanFaces', { model: 'fast' })
     * if (plugin == null) throw new Error("Failed to load scanFaces plugin!")
     * ```
     */
    initFrameProcessorPlugin(name: string, options?: Record<string, ParameterType>): FrameProcessorPlugin | undefined;
    /**
     * Throws the given error.
     */
    throwJSError(error: unknown): void;
    /**
     * Get the Frame Processor Runtime Worklet Context.
     *
     * This is the serial [DispatchQueue](https://developer.apple.com/documentation/dispatch/dispatchqueue)
     * / [Executor](https://developer.android.com/reference/java/util/concurrent/Executor) the
     * video/frame processor pipeline is running on.
     */
    workletContext: IWorkletContext | undefined;
}
export declare const VisionCameraProxy: TVisionCameraProxy;
/**
 * Runs the given function at the given target FPS rate.
 *
 * For example, if you want to run a heavy face detection algorithm
 * only once per second, you can use `runAtTargetFps(1, ...)` to
 * throttle it to 1 FPS.
 *
 * @param fps The target FPS rate at which the given function should be executed
 * @param func The function to execute.
 * @returns The result of the function if it was executed, or `undefined` otherwise.
 * @worklet
 * @example
 *
 * ```ts
 * const frameProcessor = useFrameProcessor((frame) => {
 *   'worklet'
 *   console.log('New Frame')
 *   runAtTargetFps(5, () => {
 *     'worklet'
 *     const faces = detectFaces(frame)
 *     console.log(`Detected a new face: ${faces[0]}`)
 *   })
 * })
 * ```
 */
export declare function runAtTargetFps<T>(fps: number, func: () => T): T | undefined;
/**
 * Runs the given function asynchronously, while keeping a strong reference to the Frame.
 *
 * For example, if you want to run a heavy face detection algorithm
 * while still drawing to the screen at 60 FPS, you can use `runAsync(...)`
 * to offload the face detection algorithm to a separate thread.
 *
 * @param frame The current Frame of the Frame Processor.
 * @param func The function to execute.
 * @worklet
 * @example
 *
 * ```ts
 * const frameProcessor = useFrameProcessor((frame) => {
 *   'worklet'
 *   console.log('New Frame')
 *   runAsync(frame, () => {
 *     'worklet'
 *     const faces = detectFaces(frame)
 *     const face = [faces0]
 *     console.log(`Detected a new face: ${face}`)
 *   })
 * })
 * ```
 */
export declare function runAsync(frame: Frame, func: () => void): void;
/**
 * A private API to wrap a Frame Processor with a ref-counting mechanism
 * @worklet
 * @internal
 */
export declare function wrapFrameProcessorWithRefCounting(frameProcessor: (frame: Frame) => void): (frame: Frame) => void;
export {};
//# sourceMappingURL=FrameProcessorPlugins.d.ts.map